# replicate.R
# -----------
# Requires:
#   library("digest")
#   rng.R



add.replicate <- function(conn, ngroup, nobs, nobs.dispersion,
                          nfixef, fixef.mean.df, fixef.scale,
                          nranef, ranef.cov.df, ranef.scale,
                          dispersion, family)
{
    rng <- take.rng(conn)
    rep <- gen.replicate(ngroup=ngroup, nobs=nobs,
                         nobs.dispersion=nobs.dispersion, nfixef=nfixef,
                         fixef.mean.df=fixef.mean.df, fixef.scale=fixef.scale,
                         nranef=nranef, ranef.cov.df=ranef.cov.df,
                         ranef.scale=ranef.scale, dispersion=dispersion,
                         family=family)
    put.rng(conn, rng)

    hash <- digest(rep, "md5")

    dbSendPreparedQuery(conn,
        paste('INSERT INTO "Replicate" ('
                    , '    "rng",'
                    , ''
                    , '    "group_count",'
                    , '    "sample_count",'
                    , '    "sample_dispersion",'
                    , ''
                    , '    "fixef_count",'
                    , '    "fixef_mean_df",'
                    , '    "fixef_scale",'
                    , ''
                    , '    "ranef_count",'
                    , '    "ranef_cov_df",'
                    , '    "ranef_scale",'
                    , ''
                    , '    "dispersion",'
                    , '    "family",'
                    , ''
                    , '    "md5_digest"'
                    , ') VALUES ('
                    , '    $rng,'
                    , '    $ngroup, $nobs, $nobs_dispersion,'
                    , '    $nfixef, $fixef_mean_df, $fixef_scale,'
                    , '    $nranef, $ranef_cov_df, $ranef_scale,'
                    , '    $dispersion, $family,'
                    , '    $hash'
                    , ')'
                    , sep='\n'),
                data.frame(rng=rng, ngroup=ngroup, nobs=nobs,
                           nobs_dispersion=nobs.dispersion, nfixef=nfixef,
                           fixef_mean_df=fixef.mean.df,
                           fixef_scale=fixef.scale, nranef=nranef,
                           ranef_cov_df=ranef.cov.df, ranef_scale=ranef.scale,
                           dispersion=dispersion, family=rep$family$family,
                           hash=hash))
    id <- dbGetQuery(conn, 'SELECT last_insert_rowid()')[[1]]
    id
}


load.replicate <- function(conn, id)
{
    data <- dbGetQuery(conn,
        sprintf(paste('SELECT'
                    , '    "rng",'
                    , ''
                    , '    "group_count",'
                    , '    "sample_count",'
                    , '    "sample_dispersion",'
                    , ''
                    , '    "fixef_count",'
                    , '    "fixef_mean_df",'
                    , '    "fixef_scale",'
                    , ''
                    , '    "ranef_count",'
                    , '    "ranef_cov_df",'
                    , '    "ranef_scale",'
                    , ''
                    , '    "dispersion",'
                    , '    "family",'
                    , ''
                    , '    "md5_digest"'
                    , 'FROM "Replicate"'
                    , 'WHERE "id" = %d'),
                id))

    with.rng(conn, data$rng, {
        rep <- gen.replicate(ngroup=data$group_count, nobs=data$sample_count,
                             nobs.dispersion=data$sample_dispersion,
                             nfixef=data$fixef_count,
                             fixef.mean.df=data$fixef_mean_df,
                             fixef.scale=data$fixef_scale,
                             nranef=data$ranef_count,
                             ranef.cov.df=data$ranef_cov_df,
                             ranef.scale=data$ranef_scale,
                             dispersion=data$dispersion,
                             family=data$family)
    })
    hash <- digest(rep, "md5")
    if (hash != data$md5_digest)
        stop("md5 digest does not match")

    rep
}


gen.replicate <- function(ngroup, nobs, nobs.dispersion,
                          nfixef, fixef.mean.df, fixef.scale,
                          nranef, ranef.cov.df, ranef.scale,
                          dispersion, family)
{
    # we need to be strict about integer vs. numeric for the md5sum to match
    ngroup <- as.integer(ngroup)
    nobs <- as.integer(nobs)
    nobs.dispersion <- as.numeric(nobs.dispersion)
    nfixef <- as.integer(nfixef)
    fixef.mean.df <- as.numeric(fixef.mean.df)
    fixef.scale <- as.numeric(fixef.scale)
    nranef <- as.integer(nranef)
    ranef.cov.df <- as.numeric(ranef.cov.df)
    ranef.scale <- as.numeric(ranef.scale)
    dispersion <- as.numeric(dispersion)

    # family
    if (is.character(family)) {
        family <- get(family, mode="function", envir=parent.frame())()
    } else if (is.function(family)) {
        family <- family()
    } else if (is.null(family$family)) {
        stop("'family' not recognized")
    }

    if (family$family %in% c("binomial", "poisson") && dispersion != 1) {
        warning(sprintf("dispersion parameter is ignored for '%s' family",
                        family$family))
    }

    # generate coefficient mean from t
    if (is.na(fixef.mean.df)) {
        fixef <- numeric(nfixef)
    } else {
        fixef <- fixef.scale * rt(nfixef, df=fixef.mean.df)
    }
    
    # generate coefficient covariance from inverse Wishart
    if (is.na(ranef.cov.df)) {
        ranef.cov <- diag(ranef.scale^2, nranef)
        ranef.cov.sqrt <- diag(ranef.scale, nranef)
    } else {
        ranef.cov <- (ranef.scale^2
                     * solve(rWishart(1, ranef.cov.df, diag(nranef))[,,1]))
        ranef.cov.sqrt <- chol(ranef.cov)
    }

    # generate coefficients
    u <- matrix(rnorm(ngroup * nranef), ngroup, nranef)
    ranef <- u %*% ranef.cov.sqrt

    # generate sampling rates
    rate.mean <- nobs / ngroup
    if (nobs.dispersion == Inf) {
        sample.rate <- rep(rate.mean, ngroup)
    } else {
        sample.rate <- rgamma(ngroup, nobs.dispersion,
                              nobs.dispersion / rate.mean)
        stopifnot(all(sample.rate > 0))
    }

    # generate group
    suppressWarnings({ # ignore warning about using Walker's alias method
        group <- sample.int(ngroup, nobs, replace=TRUE, prob=sample.rate)
    })

    # generate feature  matrices with Pr(x[i,j] = +1) = P(x[i,j] = -1) = 1/2,
    x <- matrix(sample(c(-1, +1), nobs * nfixef, replace=TRUE), nobs, nfixef)
    z <- matrix(sample(c(-1, +1), nobs * nranef, replace=TRUE), nobs, nranef)

    # compute linear predictors and generate observations
    eta <- drop(x %*% fixef) + rowSums(z * ranef[group,])
    mu <- family$linkinv(eta)

    if (family$family == "gaussian") {
        y <- rnorm(nobs, mean=mu, sd=sqrt(dispersion))
    } else if (family$family == "binomial") {
        y <- as.numeric(rbinom(nobs, 1, mu))
    } else if (family$family == "poisson") {
        y <- rpois(nobs, mu)
    } else {
        stop(sprintf("family '%s' not supported", family$family))
    }

    list(ngroup = ngroup, nobs = nobs,
         fixef = fixef, ranef = ranef, ranef.cov.sqrt = ranef.cov.sqrt,
         dispersion = dispersion, sample.rate = sample.rate,
         group = group, x = x, z = z, y.mean = mu, y = y, family=family)
}

